@@ -1,6 +1,4 @@
 # -*- coding: utf-8 -*-
-from __future__ import print_function
-
 import logging
 import os
 import re
@@ -9,16 +7,10 @@ import sys
 
 import django
 from django.conf import settings
-from django.core.exceptions import ImproperlyConfigured
-from django.core.management.base import BaseCommand, CommandError
+from django.core.management.base import BaseCommand, CommandError, SystemCheckError
 from django.core.servers.basehttp import get_internal_wsgi_application
-try:
-    from django.utils.autoreload import gen_filenames
-except ImportError:  # Django >=2.2
-    from django.utils.autoreload import get_reloader
-
-    def gen_filenames():
-        return get_reloader().watched_files()
+from django.dispatch import Signal
+from django.utils.autoreload import get_reloader
 
 try:
     if 'whitenoise.runserver_nostatic' in settings.INSTALLED_APPS:
@@ -34,12 +26,25 @@ from django_extensions.management.utils import RedirectHandler, has_ipdb, setup_
 from django_extensions.management.debug_cursor import monkey_patch_cursordebugwrapper
 
 
+runserver_plus_started = Signal()
 naiveip_re = re.compile(r"""^(?:
 (?P<addr>
     (?P<ipv4>\d{1,3}(?:\.\d{1,3}){3}) |         # IPv4 address
     (?P<ipv6>\[[a-fA-F0-9:]+\]) |               # IPv6 address
     (?P<fqdn>[a-zA-Z0-9-]+(?:\.[a-zA-Z0-9-]+)*) # FQDN
 ):)?(?P<port>\d+)$""", re.X)
+# 7-bit C1 ANSI sequences (https://stackoverflow.com/questions/14693701/how-can-i-remove-the-ansi-escape-sequences-from-a-string-in-python)
+ansi_escape = re.compile(r'''
+    \x1B  # ESC
+    (?:   # 7-bit C1 Fe (except CSI)
+        [@-Z\\-_]
+    |     # or [ for CSI, followed by a control sequence
+        \[
+        [0-?]*  # Parameter bytes
+        [ -/]*  # Intermediate bytes
+        [@-~]   # Final byte
+    )
+''', re.VERBOSE)
 DEFAULT_PORT = "8000"
 DEFAULT_POLLER_RELOADER_INTERVAL = getattr(settings, 'RUNSERVERPLUS_POLLER_RELOADER_INTERVAL', 1)
 DEFAULT_POLLER_RELOADER_TYPE = getattr(settings, 'RUNSERVERPLUS_POLLER_RELOADER_TYPE', 'auto')
@@ -47,6 +52,10 @@ DEFAULT_POLLER_RELOADER_TYPE = getattr(settings, 'RUNSERVERPLUS_POLLER_RELOADER_
 logger = logging.getLogger(__name__)
 
 
+def gen_filenames():
+    return get_reloader().watched_files()
+
+
 class Command(BaseCommand):
     help = "Starts a lightweight Web server for development."
 
@@ -56,7 +65,7 @@ class Command(BaseCommand):
     DEFAULT_KEY_EXTENSION = ".key"
 
     def add_arguments(self, parser):
-        super(Command, self).add_arguments(parser)
+        super().add_arguments(parser)
         parser.add_argument('addrport', nargs='?',
                             help='Optional port number, or ipaddr:port')
         parser.add_argument('--ipv6', '-6', action='store_true', dest='use_ipv6', default=False,
@@ -217,6 +226,18 @@ class Command(BaseCommand):
         with monkey_patch_cursordebugwrapper(print_sql=options["print_sql"], print_sql_location=options["print_sql_location"], logger=logger.info, confprefix="RUNSERVER_PLUS"):
             self.inner_run(options)
 
+    def get_handler(self, *args, **options):
+        """Return the default WSGI handler for the runner."""
+        return get_internal_wsgi_application()
+
+    def get_systemcheckerror_handler(self, error_message, error_class=Exception, **options):
+        error_message = ansi_escape.sub('', error_message)
+
+        def application(env, start_response):
+            raise error_class(error_message)
+
+        return application
+
     def inner_run(self, options):
         try:
             from werkzeug import run_simple
@@ -237,7 +258,7 @@ class Command(BaseCommand):
 
         class WSGIRequestHandler(_WSGIRequestHandler):
             def make_environ(self):
-                environ = super(WSGIRequestHandler, self).make_environ()
+                environ = super().make_environ()
                 if not options['keep_meta_shutdown_func']:
                     del environ['werkzeug.server.shutdown']
                 return environ
@@ -246,28 +267,31 @@ class Command(BaseCommand):
         use_reloader = options['use_reloader']
         open_browser = options['open_browser']
         quit_command = (sys.platform == 'win32') and 'CTRL-BREAK' or 'CONTROL-C'
-        extra_files = options['extra_files']
         reloader_interval = options['reloader_interval']
         reloader_type = options['reloader_type']
+        self.extra_files = set(options['extra_files'])
 
         self.nopin = options['nopin']
 
         if self.show_startup_messages:
             print("Performing system checks...\n")
-        if hasattr(self, 'check'):
-            self.check(display_num_errors=self.show_startup_messages)
-        else:
-            self.validate(display_num_errors=self.show_startup_messages)
+
         try:
-            self.check_migrations()
-        except ImproperlyConfigured:
-            pass
-        handler = get_internal_wsgi_application()
+            self.check(display_num_errors=self.show_startup_messages)
+        except SystemCheckError as exc:
+            self.stderr.write("SystemCheckError occurred during system checks: " + str(exc), ending="\n\n")
+            handler = self.get_systemcheckerror_handler(str(exc), error_class=type(exc), **options)
+        else:
+            handler = self.get_handler(**options)
+
+        self.check_migrations()
+
         if USE_STATICFILES:
             use_static_handler = options['use_static_handler']
             insecure_serving = options['insecure_serving']
             if use_static_handler and (settings.DEBUG or insecure_serving):
                 handler = StaticFilesHandler(handler)
+
         if options["cert_path"] or options["key_file_path"]:
             """
             OpenSSL is needed for SSL support.
@@ -298,7 +322,6 @@ class Command(BaseCommand):
                 if self.show_startup_messages:
                     print("Werkzeug version is less than 0.9, trying adhoc certificate.")
                 ssl_context = "adhoc"
-
         else:
             ssl_context = None
 
@@ -316,7 +339,10 @@ class Command(BaseCommand):
             webbrowser.open(bind_url)
 
         if use_reloader and settings.USE_I18N:
-            extra_files.extend(filter(lambda filename: str(filename).endswith('.mo'), gen_filenames()))
+            self.extra_files |= set(filter(lambda filename: str(filename).endswith('.mo'), gen_filenames()))
+
+        if getattr(settings, 'RUNSERVER_PLUS_EXTRA_FILES', []):
+            self.extra_files |= set(settings['RUNSERVER_PLUS_EXTRA_FILES'])
 
         # Werkzeug needs to be clued in its the main instance if running
         # without reloader or else it won't show key.
@@ -331,13 +357,14 @@ class Command(BaseCommand):
                 os.environ['WERKZEUG_DEBUG_PIN'] = 'off'
             handler = DebuggedApplication(handler, True)
 
+        runserver_plus_started.send(sender=self)
         run_simple(
             self.addr,
             int(self.port),
             handler,
             use_reloader=use_reloader,
             use_debugger=True,
-            extra_files=extra_files,
+            extra_files=self.extra_files,
             reloader_interval=reloader_interval,
             reloader_type=reloader_type,
             threaded=threaded,
